
#include <iostream>
#include <algorithm>
#include <vector>
#include <cmath>
#include "UniRand.h"


using namespace std;


// Implementation of Andrew's monotone chain 2D convex hull algorithm.
// Asymptotic complexity: O(n log n).
// Practical performance: 0.5-1.0 seconds for n=1000000 on a 1GHz machine.

typedef double coord_t;         // coordinate type
typedef double coord2_t;  // must be big enough to hold 2*max(|coordinate|)^2

struct Point {
	coord_t x, y;

	bool operator <(const Point &p) const {
		return x < p.x || (x == p.x && y < p.y);
	}
};

// 2D cross product of OA and OB vectors, i.e. z-component of their 3D cross product.
// Returns a positive value, if OAB makes a counter-clockwise turn,
// negative for clockwise turn, and zero if the points are collinear.
coord2_t cross(const Point &O, const Point &A, const Point &B)
{
	return (A.x - O.x) * (B.y - O.y) - (A.y - O.y) * (B.x - O.x);
}

// Returns a list of points on the convex hull in counter-clockwise order.
// Note: the last point in the returned list is the same as the first one.
void convex_hull(vector<Point> P, vector<Point>& H)
{
	int n = P.size(), k = 0;
	if (n == 1) return;

	// Sort points lexicographically
	sort(P.begin(), P.end());

	// Build lower hull
	for (int i = 0; i < n; ++i) {
		while (k >= 2 && cross(H[k-2], H[k-1], P[i]) <= 0) k--;
		H[k++] = P[i];
	}

	// Build upper hull
	for (int i = n-2, t = k+1; i >= 0; i--) {
		while (k >= t && cross(H[k-2], H[k-1], P[i]) <= 0) k--;
		H[k++] = P[i];
	}

	H.resize(k-1);
}


double polygon_area(int m, vector<Point> P)
{
    double area = 0.0;
    for (int i = 0; i < m; ++i){
       int j = (i + 1) % m;
       area += 0.5 * (P[i].x*P[j].y -  P[j].x*P[i].y);
    }
    return (area);
}

int main() {
    cout << "Start" << endl;

    const int npts = 20;
    const int niter = 1000000;
    vector<Point> points(npts);
    vector<Point> convex(npts);
    UniRand rng;

    double mean = 0.0;
    double dev = 0.0;

    cout << npts << " points, " << niter << " iterations" << endl;

    for (int k = 0; k < niter; k++) {
        int j = 0;
        while (j < npts) {
            double rx = 2.0 * rng.rand() - 1.0;
            double ry = 2.0 * rng.rand() - 1.0;
            if (rx * rx + ry * ry <= 1.0) {
                points[j].x = rx;
                points[j].y = ry;
                j++;
            }
        }
        convex.resize(npts);
        convex_hull(points, convex);
        int m = convex.size();
        double area = polygon_area(m, convex) / M_PI;
        mean += area;
        dev += area * area;
    }

    mean = mean / (double)niter;
    dev = dev / (double)niter - mean * mean;
    dev = sqrt(dev);
    cout << "mean area = " << mean << endl;
    cout << "s.d       = " << dev << endl;

    return 0;
}
